import pytest
from faker import Faker

from mex.artificial.helpers import create_faker, register_factories
from mex.artificial.identity import create_identities
from mex.common.models import EXTRACTED_MODEL_CLASSES

pytest_plugins = ("mex.common.testing.plugin",)


@pytest.fixture(name="faker")
def init_faker() -> Faker:
    """Return a fully configured faker instance."""
    faker = create_faker(["en_US"], 0)
    identities = create_identities(faker, len(EXTRACTED_MODEL_CLASSES) * 2)
    register_factories(faker, identities, 5)
    return faker



import re
from collections.abc import Sequence
from datetime import datetime
from enum import Enum
from functools import partial
from typing import Any, cast, get_args, get_origin

from annotated_types import MaxLen, MinLen
from faker import Generator as FakerFactory
from faker.providers import BaseProvider as BaseFakerProvider
from faker.providers.internet import Provider as InternetFakerProvider
from faker.providers.python import Provider as PythonFakerProvider
from pydantic import BaseModel
from pydantic.fields import FieldInfo

from mex.artificial.types import IdentityMap
from mex.common.identity import Identity
from mex.common.models import EXTRACTED_MODEL_CLASSES_BY_NAME, AnyExtractedModel
from mex.common.transform import ensure_prefix
from mex.common.types import (
    TEMPORAL_ENTITY_FORMATS_BY_PRECISION,
    UTC,
    Email,
    Identifier,
    Link,
    LinkLanguage,
    TemporalEntity,
    TemporalEntityPrecision,
    Text,
)


class RandomFieldInfo(BaseModel):
    """Randomized pick of matching inner type and patterns for a field."""

    inner_type: Any
    numerify_patterns: list[str] = []
    regex_patterns: list[str] = []


class BuilderProvider(PythonFakerProvider):
    """Faker provider that deals with interpreting pydantic model fields."""

    def min_max_for_field(self, field: FieldInfo) -> tuple[int, int]:
        """Return a min and max item count for a field."""
        if get_origin(field.annotation) is list:
            # calculate the item counts based on field annotations
            min_items = 0
            max_items = self.pyint(1, 3)
            if min_lengths := [x for x in field.metadata if isinstance(x, MinLen)]:
                min_items = min_lengths[0].min_length
                max_items += min_items  # the max should be higher than the min
            if max_lengths := [x for x in field.metadata if isinstance(x, MaxLen)]:
                max_items = max_lengths[0].max_length
        else:
            # required fields need at least 1 item, optional fields 0
            min_items = int(field.is_required())
            # a list with length 1 will be unpacked by `fix_listyness`
            max_items = 1
        return min_items, max_items

    def ensure_list(self, values: object) -> list[object]:
        """Wrap single object in list, replace None with [] and return list as is."""
        if values is None:
            return []
        if isinstance(values, list):
            return values
        return [values]

    def get_random_field_info(self, field: FieldInfo) -> RandomFieldInfo:
        """Randomly pick a matching type and patterns for a given field."""
        # determine field type and unpack unions, lists, and other types with args
        if args := get_args(field.annotation):
            # mixed types are not yet supported
            return self.random_element(
                [
                    self.get_random_field_info(field.from_annotation(type_))
                    for type_ in args
                    if type_ is not type(None)
                ]
            )
        return RandomFieldInfo(
            inner_type=field.annotation,
            numerify_patterns=[
                re.sub(r"[0-9]", "#", e)
                for e in self.ensure_list(field.examples)
                if isinstance(e, str)
            ],
            regex_patterns=[m.pattern for m in field.metadata if hasattr(m, "pattern")],
        )

    def field_value(
        self,
        field: FieldInfo,
        identity: Identity,
    ) -> list[Any]:
        """Get a single artificial value for the given field and identity."""
        field_info = self.get_random_field_info(field)
        if field_info.regex_patterns and field_info.numerify_patterns:
            factory = partial(
                self.generator.numerify_patterns,
                field_info.numerify_patterns,
                field_info.regex_patterns,
            )
        elif issubclass(field_info.inner_type, Identifier):
            factory = partial(
                self.generator.reference, field_info.inner_type, exclude=identity
            )
        elif issubclass(field_info.inner_type, Link):
            factory = self.generator.link
        elif issubclass(field_info.inner_type, Email):
            factory = self.generator.email
        elif issubclass(field_info.inner_type, Text):
            factory = self.generator.text_object
        elif issubclass(field_info.inner_type, TemporalEntity):
            factory = partial(
                self.generator.temporal_entity,
                field_info.inner_type.ALLOWED_PRECISION_LEVELS,
            )
        elif issubclass(field_info.inner_type, Enum):
            factory = partial(self.random_element, field_info.inner_type)
        elif issubclass(field_info.inner_type, str):
            factory = self.generator.text_string
        elif issubclass(field_info.inner_type, int):
            factory = self.generator.random_int
        else:
            msg = f"Cannot create fake data for {field}"
            raise RuntimeError(msg)
        return [
            value
            for _ in range(self.pyint(*self.min_max_for_field(field)))
            if (value := factory()) is not None
        ]

    def extracted_items(self, stem_types: Sequence[str]) -> list[AnyExtractedModel]:
        """Get a list of extracted items for the given model classes."""
        items = []
        for stem_type in stem_types:
            entity_type = ensure_prefix(stem_type, "Extracted")
            model = EXTRACTED_MODEL_CLASSES_BY_NAME[entity_type]
            for identity in cast("list[Identity]", self.generator.identities(model)):
                # manually set identity related fields
                payload: dict[str, Any] = {
                    "identifier": identity.identifier,
                    "hadPrimarySource": identity.hadPrimarySource,
                    "identifierInPrimarySource": identity.identifierInPrimarySource,
                    "stableTargetId": identity.stableTargetId,
                    "entityType": entity_type,
                }
                # dynamically populate all other fields
                for name, field in model.model_fields.items():
                    if name not in payload:
                        payload[name] = self.field_value(field, identity)
                items.append(model.model_validate(payload))
        return items


class IdentityProvider(BaseFakerProvider):
    """Faker provider that creates identities and helps with referencing them."""

    def __init__(self, factory: FakerFactory, identities: IdentityMap) -> None:
        """Create and persist identities for all entity types."""
        super().__init__(factory)
        self._identities = identities

    def identities(self, model: type[AnyExtractedModel]) -> list[Identity]:
        """Return a list of identities for the given model class."""
        return self._identities[model.__name__.removeprefix("Extracted")]

    def reference(
        self,
        inner_type: type[Identifier],
        exclude: Identity,
    ) -> Identifier | None:
        """Return ID for random identity of given type (that is not excluded)."""
        if choices := [
            identity
            for entity_type in re.findall(
                r"Merged([A-Za-z]+)Identifier", str(inner_type)
            )
            for identity in self._identities[entity_type]
            # avoid self-references by skipping excluded ids
            if identity.stableTargetId != exclude.stableTargetId
        ]:
            return Identifier(self.random_element(choices).stableTargetId)
        return None


class LinkProvider(InternetFakerProvider, PythonFakerProvider):
    """Faker provider that can return links with optional title and language."""

    def link(self) -> Link:
        """Return a link with optional title and language."""
        title, language = None, None
        if self.pybool():
            title = self.domain_word().replace("-", " ").title()
            if self.pybool():
                language = self.random_element(LinkLanguage)
        return Link(url=self.url(), title=title, language=language)


class TemporalEntityProvider(PythonFakerProvider):
    """Faker provider that can return a custom TemporalEntity with random precision."""

    def temporal_entity(
        self, allowed_precision_levels: list[TemporalEntityPrecision]
    ) -> TemporalEntity:
        """Return a custom temporal entity with random date, time and precision."""
        return TemporalEntity(
            datetime.fromtimestamp(
                self.pyint(int(8e8), int(datetime.now(tz=UTC).timestamp())), tz=UTC
            ).strftime(
                TEMPORAL_ENTITY_FORMATS_BY_PRECISION[
                    self.random_element(allowed_precision_levels)
                ]
            )
        )


class TextProvider(PythonFakerProvider):
    """Faker provider that handles custom text related requirements."""

    def __init__(self, factory: FakerFactory, chattiness: int) -> None:
        """Configure the chattiness of generated text."""
        super().__init__(factory)
        self._chattiness = chattiness

    def text_string(self) -> str:
        """Return a randomized sequence of words as a string."""
        return " ".join(self.generator.word() for _ in range(self.pyint(1, 3)))

    def text_object(self) -> Text:
        """Return a random text paragraph with an auto-detected language."""
        return Text(value=self.generator.paragraph(self.pyint(1, self._chattiness)))


class NumerifyPatternsProvider(PythonFakerProvider):
    """Faker provider that tries to numerify a pattern until it matches a regex."""

    def numerify_patterns(
        self,
        numerify_patterns: list[str],
        regex_patterns: list[str],
    ) -> str | None:
        """Try to numerify a pattern in 10 turns until it validates, or bail out."""
        for _ in range(10):
            numerify_pattern = self.random_element(numerify_patterns)
            regex_pattern = self.random_element(regex_patterns)
            numerified = self.numerify(numerify_pattern)
            if re.match(regex_pattern, numerified):
                return numerified
        return None



from re import Pattern
from typing import Annotated, Any
from unittest.mock import Mock

import pytest
from faker import Faker
from pydantic import BaseModel, Field
from pydantic.fields import FieldInfo

from mex.artificial.provider import RandomFieldInfo
from mex.common.models import ExtractedPrimarySource
from mex.common.testing import Joker
from mex.common.types import (
    APIType,
    Email,
    Identifier,
    Link,
    LinkLanguage,
    MergedPrimarySourceIdentifier,
    TemporalEntity,
    TemporalEntityPrecision,
    Text,
    TextLanguage,
)


class DummyModel(BaseModel):
    no_min: list[str] = []
    has_min: list[bytes] = Field([], min_length=2)
    has_max: list[bytes] = Field([], max_length=5)
    is_required: int
    is_optional: bool | None = None
    is_union: float | list[float]
    is_inner_union: list[float | int] = []
    is_union_with_pattern: (
        Annotated[
            str,
            Field(
                pattern=r"^https://wikidata\.org/entity/[PQ0-9]{2,64}$",
                examples=[
                    "http://wikidata.org/entity/Q679041",
                    "http://wikidata.org/entity/P123",
                ],
            ),
        ]
        | None
    ) = None
    is_nested_pattern: list[
        Annotated[
            str,
            Field(
                pattern=r"^https://dfg\.de/foobar/[0-9]{1,64}$",
                examples=["https://dfg.de/foobar/10179"],
            ),
        ]
    ] = []


def test_builder_provider_min_max_for_field(faker: Faker) -> None:
    min_max = {
        name: faker.min_max_for_field(field)
        for name, field in DummyModel.model_fields.items()
    }
    assert min_max == {
        "has_max": (0, 5),
        "has_min": (2, 3),
        "is_inner_union": (0, 1),
        "is_nested_pattern": (0, 3),
        "is_optional": (0, 1),
        "is_required": (1, 1),
        "is_union": (1, 1),
        "is_union_with_pattern": (0, 1),
        "no_min": (0, 3),
    }


def test_builder_provider_get_random_field_info(faker: Faker) -> None:
    random_field_info = {
        name: faker.get_random_field_info(field)
        for name, field in DummyModel.model_fields.items()
    }
    assert random_field_info == {
        "no_min": RandomFieldInfo(inner_type=str),
        "has_min": RandomFieldInfo(inner_type=bytes),
        "has_max": RandomFieldInfo(inner_type=bytes),
        "is_required": RandomFieldInfo(inner_type=int),
        "is_optional": RandomFieldInfo(inner_type=bool),
        "is_union": RandomFieldInfo(inner_type=float),
        "is_inner_union": RandomFieldInfo(inner_type=int),
        "is_union_with_pattern": RandomFieldInfo(
            inner_type=str,
            numerify_patterns=[
                "http://wikidata.org/entity/Q######",
                "http://wikidata.org/entity/P###",
            ],
            regex_patterns=[r"^https://wikidata\.org/entity/[PQ0-9]{2,64}$"],
        ),
        "is_nested_pattern": RandomFieldInfo(
            inner_type=str,
            numerify_patterns=["https://dfg.de/foobar/#####"],
            regex_patterns=[r"^https://dfg\.de/foobar/[0-9]{1,64}$"],
        ),
    }


@pytest.mark.parametrize(
    ("annotation", "expected"),
    [
        (
            Link,
            [
                Link(
                    language=LinkLanguage.DE, title="Pratt", url="https://turner.com/"
                ),
            ],
        ),
        (Email, ["lindathomas@example.net"]),
        (
            Text,
            [
                Text(
                    value="Language ball floor meet usually board necessary.",
                    language=TextLanguage.EN,
                )
            ],
        ),
        (TemporalEntity, [TemporalEntity("2003-06-21T16:11:41Z")]),
        (APIType, [APIType["SOAP"]]),
        (
            Annotated[
                Pattern,
                Field(
                    pattern=r"^https://foo\.example/[a-z0-9]{9}$",
                    examples=["https://foo.example/01k5qnb77"],
                ),
            ],
            ["https://foo.example/87k1qnb15"],
        ),
        (
            Annotated[
                str,
                Field(
                    pattern=r"^http://bar\.batz/[A-Z0-9]{2,64}$",
                    examples=["http://bar.batz/D001604"],
                ),
            ],
            ["http://bar.batz/D871158"],
        ),
        (
            str,
            ["suggest"],
        ),
    ],
)
def test_builder_provider_field_value(
    faker: Faker,
    annotation: Any,  # noqa: ANN401
    expected: Any,  # noqa: ANN401
) -> None:
    field = FieldInfo.from_annotation(annotation)
    identity = Mock(stableTargetId="00000000001234")

    assert faker.field_value(field, identity) == expected


def test_builder_provider_field_value_reference(faker: Faker) -> None:
    field = FieldInfo.from_annotation(MergedPrimarySourceIdentifier)
    identity = Mock(stableTargetId="00000000001234")
    reference = faker.field_value(field, identity)

    assert set(reference) < {
        i.stableTargetId for i in faker.identities(ExtractedPrimarySource)
    }


def test_builder_provider_field_value_error(faker: Faker) -> None:
    field = FieldInfo.from_annotation(object)
    identity = Mock(stableTargetId="00000000001234")

    with pytest.raises(RuntimeError, match="Cannot create fake data"):
        faker.field_value(field, identity)


def test_builder_provider_extracted_items(faker: Faker) -> None:
    models = faker.extracted_items(["ContactPoint"])
    assert models[0].model_dump(exclude_defaults=True) == {
        "email": ["udavis@example.net"],
        "hadPrimarySource": Joker(),
        "identifier": Joker(),
        "identifierInPrimarySource": "ContactPoint-4181830114",
        "stableTargetId": Joker(),
    }


def test_identity_provider_identities(faker: Faker) -> None:
    primary_sources = faker.identities(ExtractedPrimarySource)
    assert len(primary_sources) == 2
    assert primary_sources[0].model_dump() == {
        "hadPrimarySource": MergedPrimarySourceIdentifier("00000000000000"),
        "identifier": Joker(),
        "identifierInPrimarySource": "PrimarySource-2516530558",
        "stableTargetId": Joker(),
    }


def test_identity_provider_reference(faker: Faker) -> None:
    identities = list(faker.identities(ExtractedPrimarySource))

    for identity in identities:
        reference = faker.reference(MergedPrimarySourceIdentifier, identity)
        assert reference != identity.stableTargetId
        assert reference in [i.stableTargetId for i in identities]

    assert faker.reference([], Identifier("00000000000000")) is None


def test_link_provider(faker: Faker) -> None:
    assert faker.link() == Link(language=None, title=None, url="http://www.pratt.com/")


def test_temporal_entity_provider(faker: Faker) -> None:
    assert faker.temporal_entity([TemporalEntityPrecision.DAY]) == TemporalEntity(
        "2014-08-30"
    )


def test_text_provider_string(faker: Faker) -> None:
    assert faker.text_string() == "doctor edge suggest"


def test_text_provider_text(faker: Faker) -> None:
    assert faker.text_object() == Text(
        value=(
            "Sound central myself before year. Your majority feeling fact by four two. "
            "White owner onto knowledge other. First drug contain start almost wonder."
        ),
        language=TextLanguage.EN,
    )



from pathlib import Path
from typing import Annotated

import typer

from mex.artificial.helpers import generate_artificial_merged_items, write_merged_items
from mex.common.logging import logger
from mex.common.models import EXTRACTED_MODEL_CLASSES

DEFAULT_LOCALE = [
    "de_DE",
    "en_US",
]
DEFAULT_MODELS = [
    "AccessPlatform",
    "Activity",
    "BibliographicResource",
    "ContactPoint",
    "Distribution",
    "Organization",
    "OrganizationalUnit",
    "Person",
    "Resource",
    "Variable",
    "VariableGroup",
]


def artificial(  # noqa: PLR0913
    count: Annotated[
        int,
        typer.Option(
            help="Number of artificial items to approximately create.",
            min=len(EXTRACTED_MODEL_CLASSES) * 2,
            max=int(10e6 - 1),
        ),
    ] = 100,
    chattiness: Annotated[
        int,
        typer.Option(
            help="Maximum number of words to approximately produce for text fields.",
            min=2,
            max=100,
        ),
    ] = 10,
    seed: Annotated[
        int,
        typer.Option(
            help="The seed value for faker randomness.",
        ),
    ] = 0,
    locale: Annotated[
        list[str] | None,
        typer.Option(
            help="The locale to use for faker.",
            show_default=DEFAULT_LOCALE,
        ),
    ] = None,
    models: Annotated[
        list[str] | None,
        typer.Option(
            help="The names of models to use for faker.",
            show_default=DEFAULT_MODELS,
        ),
    ] = None,
    path: Annotated[
        Path | None,
        typer.Option(
            help="Where to write the resulting `publisher.ndjson` file.",
            show_default=str(Path.cwd()),
            file_okay=False,
            dir_okay=True,
            writable=True,
            resolve_path=True,
        ),
    ] = None,
) -> None:  # pragma: no cover
    """Generate merged artificial items."""
    logger.info("starting artificial data generation")
    path = path or Path.cwd()
    locale = locale or DEFAULT_LOCALE
    models = models or DEFAULT_MODELS
    items = generate_artificial_merged_items(locale, seed, count, chattiness, models)
    write_merged_items(items, path)
    logger.info("artificial data generation done")


def main() -> None:  # pragma: no cover
    """Wrap entrypoint in typer."""
    typer.run(artificial)



from pytest import LogCaptureFixture
from typer import Typer
from typer.testing import CliRunner

from mex.artificial.main import artificial


def test_main(caplog: LogCaptureFixture) -> None:
    runner = CliRunner()
    app = Typer()
    app.command()(artificial)
    result = runner.invoke(app, ["--count", "30"], catch_exceptions=False)
    assert result.exit_code == 0
    assert "artificial data generation done" in caplog.text



from binascii import crc32

from faker import Faker

from mex.artificial.types import IdentityMap
from mex.common.identity import Identity, get_provider
from mex.common.models import (
    EXTRACTED_MODEL_CLASSES,
    MEX_PRIMARY_SOURCE_STABLE_TARGET_ID,
)
from mex.common.types import MergedPrimarySourceIdentifier


def get_offset_int(cls: type) -> int:
    """Calculate an integer based on the crc32 checksum of the name of a class."""
    return crc32(cls.__name__.encode())


def create_numeric_ids(faker: Faker, count: int) -> dict[str, list[int]]:
    """Create a mapping from entity type to a list of numeric ids.

    These numeric ids can be used as seeds for the identity of artificial items.
    The seeds will be passed to `Identifier.generate(seed=...)` to get deterministic
    identifiers throughout consecutive runs of the artificial data generation.

    Args:
        faker: Instance of faker
        count: Number of ids to generate

    Returns:
        Dict with entity types and lists of numeric ids
    """
    # pick a random selection of model classes from list of extracted model classes
    choices = list(
        faker.random_choices(
            EXTRACTED_MODEL_CLASSES, length=count - len(EXTRACTED_MODEL_CLASSES)
        )
    )
    # count the picks, but use at least 2 so we can fulfill required references
    counts = {model: max(2, choices.count(model)) for model in EXTRACTED_MODEL_CLASSES}
    # build a static offset integer per class to spread out the id ranges
    offsets = {model: get_offset_int(model) for model in EXTRACTED_MODEL_CLASSES}
    # calculate numeric ids per model in the calculated quantities
    return {
        model.__name__.removeprefix("Extracted"): list(
            range(offsets[model], offsets[model] + counts[model])
        )
        for model in EXTRACTED_MODEL_CLASSES
    }


def create_identities(faker: Faker, count: int) -> IdentityMap:
    """Create the identities of the to-be-faked models.

    We do this **before** actually creating the models, because we need to be able
    to set existing stableTargetIds on reference fields.

    Args:
        faker: Instance of faker
        count: Number if identities to generate

    Returns:
        Dict with entity types and lists of Identities
    """
    # create the numeric id dictionary
    numeric_ids = create_numeric_ids(faker, count)
    # store a map of identities by entity type
    identity_map: IdentityMap = {entity_type: [] for entity_type in numeric_ids}
    # primary sources need identities first
    identity_provider = get_provider()
    for entity_type in sorted(identity_map, key=lambda t: t != "PrimarySource"):
        for numeric_id in numeric_ids[entity_type]:
            # create a persistent identity for each numeric id
            if primary_sources := identity_map["PrimarySource"]:
                # pick a random primary source to "extract" the identity from
                primary_source: Identity = faker.random_element(primary_sources)
                primary_source_id = MergedPrimarySourceIdentifier(
                    primary_source.stableTargetId
                )
            else:
                # the first primary source is extracted from this placeholder ID
                primary_source_id = MEX_PRIMARY_SOURCE_STABLE_TARGET_ID
            identity = identity_provider.assign(
                primary_source_id, f"{entity_type}-{numeric_id}"
            )
            identity_map[entity_type].append(identity)
    return identity_map



from itertools import pairwise

from faker import Faker

from mex.artificial.identity import (
    create_identities,
    create_numeric_ids,
    get_offset_int,
)
from mex.common.models import EXTRACTED_MODEL_CLASSES
from mex.common.testing import Joker


def test_create_numeric_ids(faker: Faker) -> None:
    numeric_ids = create_numeric_ids(faker, 10)
    assert numeric_ids == {
        "AccessPlatform": [3427526317, 3427526318],
        "Activity": [2405477151, 2405477152],
        "BibliographicResource": [2035724763, 2035724764],
        "Consent": [2933388007, 2933388008],
        "ContactPoint": [4181830114, 4181830115],
        "Distribution": [1589096615, 1589096616],
        "Organization": [991028314, 991028315],
        "OrganizationalUnit": [3284134756, 3284134757],
        "Person": [639677827, 639677828],
        "PrimarySource": [2516530558, 2516530559],
        "Resource": [2676315731, 2676315732],
        "Variable": [4015597000, 4015597001],
        "VariableGroup": [1476619723, 1476619724],
    }


def test_get_offset_int() -> None:
    offsets = [get_offset_int(model) for model in EXTRACTED_MODEL_CLASSES]
    assert offsets[0] == 3427526317  # sanity check

    # ensure the distance between the offsets is greater than the allowed total
    min_distance = min([abs(j - i) for i, j in pairwise(offsets)])
    max_allowed = int(10e6 - 1)
    assert min_distance > max_allowed


def test_create_identities(faker: Faker) -> None:
    identity_map = create_identities(faker, 10)
    assert len(identity_map) == len(EXTRACTED_MODEL_CLASSES)
    assert identity_map["PrimarySource"][0].model_dump() == {
        "identifier": Joker(),
        "hadPrimarySource": "00000000000000",
        "identifierInPrimarySource": "PrimarySource-2516530558",
        "stableTargetId": Joker(),
    }



import json
from collections.abc import Iterable, Sequence
from os import PathLike
from pathlib import Path
from typing import cast

from faker import Faker
from faker.typing import SeedType

from mex.artificial.identity import create_identities
from mex.artificial.provider import (
    BuilderProvider,
    IdentityProvider,
    LinkProvider,
    NumerifyPatternsProvider,
    TemporalEntityProvider,
    TextProvider,
)
from mex.artificial.types import IdentityMap, LocaleType
from mex.common.logging import logger
from mex.common.merged.main import create_merged_item
from mex.common.models import AnyExtractedModel, AnyMergedModel
from mex.common.transform import MExEncoder


def create_faker(locale: LocaleType | list[str], seed: SeedType) -> Faker:
    """Create and initialize a new faker instance with the given locale and seed."""
    faker = Faker(locale=locale)
    faker.seed_instance(seed=seed)
    return faker


def register_factories(faker: Faker, identities: IdentityMap, chattiness: int) -> None:
    """Create faker providers and register them on each factory."""
    for factory in faker.factories:
        factory.add_provider(IdentityProvider(factory, identities))
        factory.add_provider(LinkProvider(factory))
        factory.add_provider(NumerifyPatternsProvider(factory))
        factory.add_provider(BuilderProvider(factory))
        factory.add_provider(TextProvider(factory, chattiness))
        factory.add_provider(TemporalEntityProvider(factory))


def create_merged_items(
    extracted_items: list[AnyExtractedModel],
) -> list[AnyMergedModel]:
    """Create merged items for a list of extracted items."""
    return [
        create_merged_item(m.stableTargetId, [m], None, validate_cardinality=True)
        for m in extracted_items
    ]


def generate_artificial_extracted_items(
    locale: LocaleType,
    seed: SeedType,
    count: int,
    chattiness: int,
    stem_types: Sequence[str],
) -> list[AnyExtractedModel]:
    """Generate a list of artificial extracted items for the given settings."""
    faker = create_faker(locale, seed)
    identities = create_identities(faker, count)
    register_factories(faker, identities, chattiness)
    return cast("list[AnyExtractedModel]", faker.extracted_items(stem_types))


def generate_artificial_merged_items(
    locale: LocaleType,
    seed: SeedType,
    count: int,
    chattiness: int,
    stem_types: Sequence[str],
) -> list[AnyMergedModel]:
    """Generate a list of artificial merged items for the given settings."""
    extracted_items = generate_artificial_extracted_items(
        locale, seed, count, chattiness, stem_types
    )
    return create_merged_items(extracted_items)


def write_merged_items(
    items: Iterable[AnyMergedModel], out_path: PathLike[str]
) -> None:
    """Write the incoming items into a new-line delimited JSON file."""
    logging_counter = 0
    with (Path(out_path) / "publisher.ndjson").open("w", encoding="utf-8") as fh:
        for item in items:
            line = json.dumps(item, ensure_ascii=False, sort_keys=True, cls=MExEncoder)
            fh.write(f"{line}\n")
            logging_counter += 1
    logger.info("%s merged items were written", logging_counter)
